<?php
namespace EpTest\Http\Middleware;

use Ep\Cache\AbstractCacheFactory;
use Ep\Http\AbstractHttpClientFactory;
use Ep\Http\Middleware\Cache;
use GuzzleHttp\Client;
use PHPUnit\Framework\TestCase;
use Psr\Container\ContainerInterface;
use Symfony\Component\Cache\Adapter\ArrayAdapter;
use Zend\ServiceManager\ServiceManager;

/**
 * Class CacheMiddlewareTest
 *
 * @package EpTest\Http\Middleware
 * @author Elektronika Perhoč <support@ep.hr>
 * @link https://bitbucket.org/elektronika-perhoc/ep-framework
 */
class CacheTest extends TestCase
{
    /** @var ContainerInterface */
    private $container;
    /** @var Client */
    private $client;
    /** @var ArrayAdapter */
    private $cache;

    protected function setUp()
    {
        parent::setUp();

        $this->container = new ServiceManager([
            'abstract_factories' => [
                AbstractCacheFactory::class,
                AbstractHttpClientFactory::class,
            ],
        ]);

        $this->container->setService('config', [
            'http_clients' => [
                'http.cached' => [
                    'base_uri' => 'http://httpbin.org/',
                    'http_errors' => false,
                    'cache' => 'cache.http',
                ],

                'http.nullcached' => [
                    'base_uri' => 'http://httpbin.org/',
                    'http_errors' => false,
                    'cache' => 'cache.null',
                ],
            ],

            'caches' => [
                'cache.http' => [
                    'adapter' => 'array',
                    'options' => [
                        'default_lifetime' => 30,
                    ],
                ],

                'cache.null' => [
                    'adapter' => 'null',
                ],
            ],
        ]);

        $this->client = $this->container->get('http.cached');
        $this->cache = $this->container->get('cache.http');

    }

    public function testResponsesHaveCacheInfoHeaders()
    {
        $response = $this->client->get('get');

        $this->assertTrue($response->hasHeader(Cache::HEADER_CACHE_INFO));
        $this->assertEquals(Cache::HEADER_CACHE_MISS, $response->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $cachedResponse = $this->client->get('/get');

        $this->assertTrue($response->hasHeader(Cache::HEADER_CACHE_INFO));
        $this->assertEquals(Cache::HEADER_CACHE_HIT, $cachedResponse->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $this->assertCount(1, array_filter($this->cache->getValues()));
    }

    public function testErrorReponsesAreNotCached()
    {
        $response = $this->client->get('status/500');

        $this->assertTrue($response->hasHeader(Cache::HEADER_CACHE_INFO));
        $this->assertEquals(Cache::HEADER_CACHE_MISS, $response->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $cachedResponse = $this->client->get('status/500');

        $this->assertTrue($response->hasHeader(Cache::HEADER_CACHE_INFO));
        $this->assertEquals(Cache::HEADER_CACHE_MISS, $cachedResponse->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $this->assertCount(0, array_filter($this->cache->getValues()));
    }

    public function testResponseMaxAgeIsUsed()
    {
        $response = $this->client->get('cache/1');

        $this->assertEquals(Cache::HEADER_CACHE_MISS, $response->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $cachedResponse = $this->client->get('cache/1');

        $this->assertEquals(Cache::HEADER_CACHE_HIT, $cachedResponse->getHeaderLine(Cache::HEADER_CACHE_INFO));

        sleep(2);

        $staleResponse = $this->client->get('cache/1');
        $this->assertEquals(Cache::HEADER_CACHE_MISS, $staleResponse->getHeaderLine(Cache::HEADER_CACHE_INFO));
    }

    public function testNotCachedIfMaxAgeIsZero()
    {
        $response = $this->client->get('cache/0');

        $this->assertEquals(Cache::HEADER_CACHE_MISS, $response->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $cachedResponse = $this->client->get('cache/0');
        $this->assertEquals(Cache::HEADER_CACHE_MISS, $cachedResponse->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $staleResponse = $this->client->get('cache/0');
        $this->assertEquals(Cache::HEADER_CACHE_MISS, $staleResponse->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $this->assertCount(0, array_filter($this->cache->getValues()));
    }

    public function testWillReturnProperResponseOnCacheMiss()
    {
        $response = $this->client->get('anything');
        $content = $response->getBody()->getContents();

        $this->assertEquals('application/json', $response->getHeaderLine('Content-Type'));
        $this->assertNotEmpty($content);

        $decoded = json_decode($content);

        $this->assertNotNull($decoded);
    }

    public function testWillNotCacheIfNoCacheControlHeaders()
    {
        $response = $this->client->get('response-headers?Cache-Control=no-cache');

        $this->assertEquals(Cache::HEADER_CACHE_MISS, $response->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $cachedResponse = $this->client->get('response-headers?Cache-Control=no-cache');
        $this->assertEquals(Cache::HEADER_CACHE_MISS, $cachedResponse->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $staleResponse = $this->client->get('response-headers?Cache-Control=no-cache');
        $this->assertEquals(Cache::HEADER_CACHE_MISS, $staleResponse->getHeaderLine(Cache::HEADER_CACHE_INFO));

        $this->assertCount(0, array_filter($this->cache->getValues()));
    }
}